@@ -88,6 +88,7 @@ module Data.Enumerator.Text
 	-- ** Unsorted
 	, require
 	, isolate
+	, isolateWhile
 	, splitWhen
 	, lines
 	
@@ -129,9 +130,8 @@ import qualified System.IO as IO
 import           System.IO.Error (isEOFError)
 import           System.IO.Unsafe (unsafePerformIO)
 
-import           Data.Enumerator hiding ( head, drop, generateM, filterM, consume
-                                        , concatMapM, iterateM, repeatM, replicateM
-                                        , foldM)
+import           Data.Enumerator.Internal
+import           Data.Enumerator (isEOF, tryIO, throwError)
 import qualified Data.Enumerator.List as EL
 import           Data.Enumerator.Util (tSpanBy, tlSpanBy, reprWord, reprChar, textToStrict)
 
@@ -216,7 +216,7 @@ concatMapM f = checkDone (continue . step) where
 	loop k (x:xs) = do
 		fx <- lift (f x)
 		k (Chunks [fx]) >>==
-			checkDoneEx (Chunks [T.pack xs]) (\k' -> loop k' xs)
+			checkDoneEx (Chunks [T.pack xs]) (`loop` xs)
 
 -- | Similar to 'Data.Enumerator.Text.concatMap', but with a stateful step
 -- function.
@@ -347,7 +347,7 @@ take n = continue (loop id n) where
 		len = toInteger (TL.length lazy)
 		
 		iter = if len < n'
-			then continue (loop (acc . (TL.append lazy)) (n' - len))
+			then continue (loop (acc . TL.append lazy) (n' - len))
 			else let
 				(xs', extra) = TL.splitAt (fromInteger n') lazy
 				in yield (acc xs') (toChunks extra)
@@ -364,7 +364,7 @@ takeWhile p = continue (loop id) where
 		lazy = TL.fromChunks xs
 		(xs', extra) = tlSpanBy p lazy
 		iter = if TL.null extra
-			then continue (loop (acc . (TL.append lazy)))
+			then continue (loop (acc . TL.append lazy))
 			else yield (acc xs') (toChunks extra)
 	loop acc EOF = yield (acc TL.empty) EOF
 
@@ -376,7 +376,7 @@ consume = continue (loop id) where
 	loop acc (Chunks []) = continue (loop acc)
 	loop acc (Chunks xs) = iter where
 		lazy = TL.fromChunks xs
-		iter = continue (loop (acc . (TL.append lazy)))
+		iter = continue (loop (acc . TL.append lazy))
 	loop acc EOF = yield (acc TL.empty) EOF
 
 -- | Pass input from a stream through two iteratees at once. Excess input is
@@ -686,7 +686,7 @@ require n = continue (loop id n) where
 		lazy = TL.fromChunks xs
 		len = toInteger (TL.length lazy)
 		iter = if len < n'
-			then continue (loop (acc . (TL.append lazy)) (n' - len))
+			then continue (loop (acc . TL.append lazy) (n' - len))
 			else yield () (toChunks (acc lazy))
 	loop _ _ EOF = throwError (Exc.ErrorCall "require: Unexpected EOF")
 
@@ -707,10 +707,27 @@ isolate n (Continue k) = continue loop where
 			then k (Chunks xs) >>== isolate (n - len)
 			else let
 				(s1, s2) = TL.splitAt (fromInteger n) lazy
-				in k (toChunks s1) >>== (\step -> yield step (toChunks s2))
-	loop EOF = k EOF >>== (\step -> yield step EOF)
+				in k (toChunks s1) >>== (`yield` toChunks s2)
+	loop EOF = k EOF >>== (`yield` EOF)
 isolate n step = drop n >> return step
 
+-- | @'isolateWhile' p@ reads characters from the stream until /p/ is false, and
+-- passes them to its iteratee. If the iteratee finishes early, characters
+-- continue to be consumed from the outer stream until /p/ is false.
+--
+-- Since: 0.4.16
+isolateWhile :: Monad m => (Char -> Bool) -> Enumeratee T.Text T.Text m b
+isolateWhile p (Continue k) = continue loop where
+	loop (Chunks []) = continue loop
+	loop (Chunks xs) = iter where
+		lazy = TL.fromChunks xs
+		(s1, s2) = TL.span p lazy
+		iter = if TL.null s2
+			then k (Chunks xs) >>== isolateWhile p
+			else k (toChunks s1) >>== (`yield` toChunks s2)
+	loop EOF = k EOF >>== (`yield` EOF)
+isolateWhile p step = Data.Enumerator.Text.dropWhile p >> return step
+
 -- | Split on characters satisfying a given predicate.
 --
 -- Since: 0.4.8
@@ -838,13 +855,18 @@ encode codec = checkDone (continue . step) where
 decode :: Monad m => Codec
        -> Enumeratee B.ByteString T.Text m b
 decode codec = checkDone (continue . step B.empty) where
-	step _   k EOF = yield (Continue k) EOF
+	step acc k EOF = if B.null acc
+		then yield (Continue k) EOF
+		else throwError (Exc.ErrorCall "Unexpected EOF while decoding")
 	step acc k (Chunks xs) = loop acc k xs
 	
 	loop acc k [] = continue (step acc k)
 	loop acc k (x:xs) = let
 		(text, extra) = codecDecode codec (B.append acc x)
-		extraChunks = Chunks (either snd id extra : xs)
+		extraChunks = Chunks $ case extra of
+			Right bytes | B.null bytes -> xs
+			Right bytes -> bytes:xs
+			Left (_, bytes) -> bytes:xs
 		
 		checkError k' = case extra of
 			Left (exc, _) -> throwError exc
@@ -927,8 +949,8 @@ utf16_le = Codec name enc dec where
 		       | (n + 1) == maxN = decodeTo n
 		loop n = let
 			req = utf16Required
-				(B.index bytes 0)
-				(B.index bytes 1)
+				(B.index bytes n)
+				(B.index bytes (n + 1))
 			decodeMore = loop $! n + req
 			in if n + req > maxN
 				then decodeTo n
@@ -952,8 +974,8 @@ utf16_be = Codec name enc dec where
 		       | (n + 1) == maxN = decodeTo n
 		loop n = let
 			req = utf16Required
-				(B.index bytes 1)
-				(B.index bytes 0)
+				(B.index bytes (n + 1))
+				(B.index bytes n)
 			decodeMore = loop $! n + req
 			in if n + req > maxN
 				then decodeTo n
